package edu.uoc.pfc.sm.servlet;

import edu.uoc.pfc.sm.annotation.ErrorDestination;
import edu.uoc.pfc.sm.annotation.Forward;
import edu.uoc.pfc.sm.annotation.NoForwardNoRedirect;
import edu.uoc.pfc.sm.annotation.Redirect;
import edu.uoc.pfc.sm.annotation.Request;
import edu.uoc.pfc.sm.annotation.Response;
import edu.uoc.pfc.sm.annotation.RolesAllowed;
import edu.uoc.pfc.sm.annotation.SMControllerBean;
import edu.uoc.pfc.sm.data.Data;
import edu.uoc.pfc.sm.error.ErrorActionMethodExecution;
import edu.uoc.pfc.sm.error.ErrorActionMethodNotAllowed;
import edu.uoc.pfc.sm.error.ErrorActionMultipleForwardRedirect;
import edu.uoc.pfc.sm.error.ErrorActionNewInstance;
import edu.uoc.pfc.sm.error.ErrorActionNotAllowed;
import edu.uoc.pfc.sm.error.ErrorSMControllerValidator;
import edu.uoc.pfc.sm.error.SMControllerError;
import edu.uoc.pfc.sm.error.SMControllerErrorHandler;
import edu.uoc.pfc.sm.util.Constants;
import edu.uoc.pfc.sm.validation.SMControllerValidator;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.apache.commons.beanutils.BeanUtilsBean;
import org.apache.commons.beanutils.BeanUtilsBean2;

/**
 * Servlet que implementa el controlador del SMFramework
 * 
 * @author Salvador Morla
 * @version 1.0
 */
public class SMController extends HttpServlet {

    /* Determina el paquet base de l'aplicació */
    String  basePackageName;
    /* Determina el directori a on se troben les vistes de l'aplicació */
    String  applicationViewFolder;
    /* Indica si està o no activat el mode debug */
    boolean debug;
    /* Component de gestió d'errors per al controlador */
    SMControllerErrorHandler errorHandler;
    /* Component de gestió de validació de dades del controlador */
    SMControllerValidator controllerValidator;
            
    /**
     * Mètode d'inicialització del Servlet que implementa el controlador.
     * 
     * Se determinen els paràmetres d'inicialització del controlador:
     * - basePackageName
     * - encoding
     * - applicationViewFolder
     * - debug
     */
    @Override
    public void init(ServletConfig config) throws ServletException {
        
        System.out.println("SMController 1.0");
        
        // Configuram la variable que estableix el package per les classes action
        this.basePackageName = config.getInitParameter("basePackageName");
        if (this.basePackageName == null || this.basePackageName.isEmpty()) {
            throw new ServletException("SMController mal configurat. S'ha d'especificar el paquet bàse de l'aplicació.");
        }
        System.out.println("SMController: base package configurat a " + basePackageName + ".");
        
        this.applicationViewFolder = config.getInitParameter("applicationViewFolder");
        if (this.applicationViewFolder == null || this.applicationViewFolder.isEmpty()) {
            throw new ServletException("SMController mal configurat. S'ha d'especificar el directori de vistes de l'aplicació.");
        }
        System.out.println("SMController: directori de vistes de l'aplicació configurat a " + applicationViewFolder + ".");
        
        // Configuram la variable que estableix si volem informació de debug
        // al processament de les peticions per SMController
        debug = (config.getInitParameter("debug") != null) && (config.getInitParameter("debug").equals("true"));
        System.out.println("SMController: la informació de debug està " + (debug ? "habilitada" : "deshabilitada") + ".");
        
        // Inicial un gestor d'errors del controlador
        this.errorHandler = new SMControllerErrorHandler();
    }

    /**
     * Processament de peticions per als mètodes HTTP de <code>GET</code> i
     * <code>POST</code>.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException Si se produeix un error de Servlet
     * @throws IOException Si se produeix una excepció de E/S
     */
    protected void processRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        
        /* Representa l'objecte instancia de classe de componet de controlador Ation */
        Object action = null;
        /* Path de la petició */
        String path;
        
        // Pattern emprat  per obtenir el nom del component de controlador Action 
        // i l'operació d'aquest que s'indica a la peticio.
        Pattern actionNamePattern    = Pattern.compile("[^/]([^/.]+)[\\.]{0,1}([^/.]*)$"); 
        // Pattern emprat per obtenir la ruta.
        Pattern actionPackagePattern = Pattern.compile("[/](.+)[/]");
        
        /* Obtenim la petició */
        path = request.getPathInfo();
        if (path == null) {
            throw new ServletException("SMController Error. SMController mal configurat. Afegir mapping a /actions/* al fitxer web.xml");
        }
        
        /**
         * Secció de la instanciació de l'acció que correspon en funció de la 
         * petició
         */
        Matcher packageMatcher = actionPackagePattern.matcher(path);
        Matcher actionMatcher  = actionNamePattern.matcher(path);
        String actionsPackageName = this.basePackageName + ".actions";

        String actionPackageName = (packageMatcher.find() ? packageMatcher.group() : "/").replaceAll("/", ".");
        String actionName = actionMatcher.find() ? actionMatcher.group() : "";
        String operationName = null;

        /*
         * Mitjançant un string tokenizer trobam el nom de l'Action i en cas
         * de que s'hagi especificat, l'operació dins l'Action que s'ha 
         * d'executar.
         */
        if (actionName.contains(".")) {
            StringTokenizer stk = new StringTokenizer(actionName, ".");
            if (stk.hasMoreTokens()) {
                actionName = stk.nextToken();
            }
            if (stk.hasMoreTokens()) {
                operationName = stk.nextToken();
            }
        }
        /* 
         * Una vegada processada la cadena que conte la petició estam en disposició
         * d'obtenir el fullyQualifiedName de la accio
         */
        String fullyQualifiedClassName = actionsPackageName + actionPackageName + actionName;
        if (operationName == null) {            
            /* la acció per defecte intentarà executar el mètode execute() */
            operationName = "execute"; 
        }
        
        /*
         * Hem determinat quina es l'acció de la petició.
         */
        debug("SMController. S'ha interceptat la petició: "+actionName + "." + operationName + "()");
        
        
        /**
         * El controlador intenta trobar l'acció indicada en la petició. 
         * Si no la troba, s'asumeix que hem de fer un forward cap a la vista
         * que te per nom la petició indicada i acaba el processament.
         */
        try {
            action = Thread.currentThread().getContextClassLoader().loadClass(fullyQualifiedClassName).newInstance();
        } catch (ClassNotFoundException ex) {            
            debug("SMController Warning. No existeix la Action " + actionName + "(" + ex.getMessage() + ")");
            String jsp = "/" + applicationViewFolder + path + ".jsp";
            debug("SMController. Redirigint la navegació a la vista (JSP) : " + jsp);
            RequestDispatcher dispatcher = request.getRequestDispatcher(jsp);
            dispatcher.forward(request, response);
            return;            
        } catch (InstantiationException ex) {
            errorHandler.handleError(request, response, null, new ErrorActionNewInstance(fullyQualifiedClassName, ex));
        } catch (IllegalAccessException ex) {
            errorHandler.handleError(request, response, null, new ErrorActionNewInstance(fullyQualifiedClassName, ex));
        }
        
        
        /**
         * Inicialització dels valors redirect i forward per defecte a nivell de 
         * classe Action. Recuperam les anotacions de l'action:
         */
        Redirect classRedirect = action.getClass().getAnnotation(Redirect.class);
        Forward classForward = action.getClass().getAnnotation(Forward.class);
        ErrorDestination classErrorDestination = action.getClass().getAnnotation(ErrorDestination.class);
        
        
        /**
         * Secció de comprobació de restriccions d'accés per rol
         */        
        RolesAllowed rolesAllowed = action.getClass().getAnnotation(RolesAllowed.class);
        debug("SMController. Processament de la seguretat a nivell de Action.");
        if (rolesAllowed != null) {
            debug("SMController. Permes l'acces pels rols:");
            String[] roles;
            roles = rolesAllowed.value();
            boolean found = false;
            for (String rol : roles) {
                debug(" - " + rol);
                found = found || request.isUserInRole(rol);
            }
            if (!found) {
                errorHandler.handleError(request, response, classErrorDestination, new ErrorActionNotAllowed(fullyQualifiedClassName));
            }
        }
                
        
        /**
        * Processament de la validació de dades d'entrada.
        * Se recorren els atributs dels SMControllerBeans i se procesen les 
        * entrades que han estat anotades amb validate.
        * La validació se realitza a nivell de metode. 
        */
        controllerValidator = new SMControllerValidator(action, operationName, request, debug);
        try {            
            controllerValidator.executeValidation();            
        }
        catch (Exception e) {
            errorHandler.handleError(request, response, classErrorDestination, new ErrorSMControllerValidator(fullyQualifiedClassName, operationName, e));
        }        
        // Si la llista s'han produit errors de validacio
        // el controlador mostra la vista especifica d'error de validació definida
        // pel mètode del component de controlador Action.
         if (controllerValidator.getError() != null && !controllerValidator.getError().getErrorList().isEmpty()) {            
            controllerValidator.getError().setError(Boolean.TRUE);            
            debug("SMController. Hi ha errors de validacio");
            debug("SMController. Redirigint la navegació a la vista que gestiona els errors de validació " + controllerValidator.getError().getValidationErrorView());
            
            // Ficam l'error al Request com un attribut:
            request.setAttribute("SMControllerValidationError", controllerValidator.getError());
            
            RequestDispatcher dispatcher = request.getRequestDispatcher(controllerValidator.getError().getValidationErrorView());
            dispatcher.forward(request, response);
            return;            
        }
        
        
        /**
         * Se recorren els atributs de la classe Action i s'inicialitzen
         * els SMControllerBean. 
         * Endemés s'injecten el Request i Response si així s'ha sol·licitat
         */
        debug("SMController. Inicialització dels SMControllerBeans.");
        try {
            Field[] declaredFields = action.getClass().getDeclaredFields();
            for (Field field : declaredFields) {
                
                // Evaluam una possible injecció de dades del request al bean
                if (field.getAnnotation(SMControllerBean.class) != null) { 
                    SMControllerBean annotation = field.getAnnotation(SMControllerBean.class);
                    
                    // Tipus bàsics: requereixen un tractament especial
                    if (annotation.create() && field.getType().isPrimitive()) {
                        
                        if (field.getType().equals(Integer.TYPE)) {
                            setFieldValue(field, action, annotation.requestGetParameter() ? getIntFromRequest(request, field.getName()) : Constants.DEFAULT_INT);
                        }
                        if (field.getType().equals(Float.TYPE)) {
                            setFieldValue(field, action, annotation.requestGetParameter() ? getFloatFromRequest(request, field.getName()) : Constants.DEFAULT_FLOAT);
                        }
                        if (field.getType().equals(Double.TYPE)) {
                            setFieldValue(field, action, annotation.requestGetParameter() ? getDoubleFromRequest(request, field.getName()) : Constants.DEFAULT_DOUBLE);
                        }
                        if (field.getType().equals(Boolean.TYPE)) {
                            setFieldValue(field, action, annotation.requestGetParameter() ? getBooleanFromRequest(request, field.getName()) : Constants.DEFAULT_BOOLEAN);
                        }
                        if (field.getType().equals(Character.TYPE)) {
                            setFieldValue(field, action, annotation.requestGetParameter() ? getCharFromRequest(request, field.getName()) : Constants.DEFAULT_CHAR);
                        }                        
                    } 
                    // Se crea i mapeja del request l'attribut 'field'
                    else if (annotation.create() && annotation.requestGetParameter()) { 
                        Object obj = getParametersFromRequest(request, field);
                        setFieldValue(field, action, obj);
                    } 
                    // Se solicita create() pero sense mapeig del request
                    else if (annotation.create()) { 
                        Object obj = createNewInstance(field.getType());
                        setFieldValue(field, action, obj);
                    }                    
                    debug("SMController. Inicialitzat bean " + field.getName() + " ("+ field.getType().getSimpleName() +") ");
                } 
                // Injecció del request
                else if (field.getAnnotation(Request.class) != null)  { 
                    setFieldValue(field, action, request);
                    debug("SMController. Request injectat ");
                } 
                // Injecció del response
                else if (field.getAnnotation(Response.class) != null) { 
                    setFieldValue(field, action, response);
                    debug("SMController. Response injectat ");
                }
            }
        } catch (Exception ex) {
            throw new ServletException(ex);
        }

        /* 
         * Execució de l'Action (i actualització de forward i redirect (si
         * procedeix) 
         */
        Redirect methodRedirect = null;
        Forward methodForward = null;
        RolesAllowed methodRolesAllowed;
        ErrorDestination methodErrorDestination = null;
        NoForwardNoRedirect noForwardNoRedirect = null;
        debug("SMController. Execució dels procediment de la classe Action.");
        try {
            /* 
             * Se fica al request la classe i metodes executats per si el JSP
             * ho necessita coneixer.
             */
            request.setAttribute("SMControllerExecution", fullyQualifiedClassName + "." + operationName + "()");
            Method m = action.getClass().getMethod(operationName, (Class<?>[]) null);

            methodForward = m.getAnnotation(Forward.class);
            methodRedirect = m.getAnnotation(Redirect.class);
            methodRolesAllowed = m.getAnnotation(RolesAllowed.class);
            methodErrorDestination = m.getAnnotation(ErrorDestination.class);
            noForwardNoRedirect = m.getAnnotation(NoForwardNoRedirect.class);

            if (methodErrorDestination == null) {
                methodErrorDestination = classErrorDestination;
            }

            if (methodRedirect != null && methodForward != null) {
                errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMultipleForwardRedirect(fullyQualifiedClassName, operationName));
            }
            
            /*
             * Comprobació de permissos a nivell de mètode
             */
            if (methodRolesAllowed != null) {
                boolean methodAllowed = false;
                for (String rol : methodRolesAllowed.value()) {
                    methodAllowed = methodAllowed || request.isUserInRole(rol);
                }
                if (!methodAllowed) {
                    errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodNotAllowed(fullyQualifiedClassName, operationName));
                }
            }
            
            // Invocació del mètode de l'Action
            m.invoke(action, (Object[]) null);
        } catch (IllegalAccessException ex) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodNotAllowed(fullyQualifiedClassName, operationName, ex));
        } catch (IllegalArgumentException ex) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodExecution(fullyQualifiedClassName, operationName, ex));
        } catch (InvocationTargetException ex) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodExecution(fullyQualifiedClassName, operationName, ex));
        } catch (NoSuchMethodException ex) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodExecution(fullyQualifiedClassName, operationName, ex));
        } catch (SecurityException ex) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMethodNotAllowed(fullyQualifiedClassName, operationName, ex));
        }


        /* 
         * Per ultim se repasen els atributs SMControllerBean de l'action i per
         * aquells que no son nuls i han estat marcats como setRequestAttribute
         * s'inserten al request per posar-los a disposició de qualsevol compo-
         * nent amb acces al Request.
         */
        Field[] declaredFields = action.getClass().getDeclaredFields();
        for (Field field : declaredFields) {
            SMControllerBean annotation = field.getAnnotation(SMControllerBean.class);
            if (annotation != null) {
                try {
                    boolean accessible = field.isAccessible();
                    field.setAccessible(true);

                    Object fieldInstance = field.get(action);
                    field.setAccessible(accessible);
                    
                    // Si estaven marcats com create() ja se van inicialitzar abans
                    // d'executar l'operació.
                    if (fieldInstance != null && annotation.requestSetAttribute()) { 
                        request.setAttribute(field.getName(), fieldInstance);
                    } else if (annotation.requestSetAttribute()) {
                        debug("No s'inserta " + field.getName() + " de la classe " + field.getType() + " porquè te valor nul.");
                    } else {
                        debug("No s'inserta " + field.getName() + " de la classe " + field.getType() + " perque esta marcat com requestSetAttribute=false a SMControllerBean.");
                    }

                } catch (Exception ex) {
                    debug("S'ha intentat mapejar " + field.getName() + " de la classe " + field.getType() + " al request, pero s'ha produit un error: " + ex.getMessage());
                }
            }
        }
        

        /*
         * Analisi del destí; se pren primer 'redirect' i llavors 'forward'.
         * Primer el definit a nivell de metode i llavors el definit a nivell 
         * de classe.
         */
        if (classRedirect != null && classForward != null) {
            errorHandler.handleError(request, response, methodErrorDestination, new ErrorActionMultipleForwardRedirect(fullyQualifiedClassName, actionName));
        } else if (noForwardNoRedirect != null && noForwardNoRedirect.value()) {
            /*
             * En aquest cas no se fa res; si s'indica l'annotació 
             * @NoForwardNoredirect a un mètode el controlador no redirigeix i 
             * fa forward (sera l'action la que ho gestioni.
             */
            debug("La Action o el metode de la Action executat s'ha marcat com NoForwardNoRedirect");
        } else if (methodRedirect != null) {
            debug("SMController. S'ha definit 'Redirect' a nivell de metode (" + methodRedirect.value() + ").");
            response.sendRedirect(request.getContextPath() + (methodRedirect.value().startsWith("/") ? methodRedirect.value() : ("/" + methodRedirect.value())));
        } else if (methodForward != null) {
            debug("SMController. S'ha definit 'Forward' a nivell de metode (" + methodForward.value() + ").");
            request.getRequestDispatcher(methodForward.value().startsWith("/") ? methodForward.value() : ("/" + methodForward.value())).forward(request, response);
        } else if (classRedirect != null) {
            debug("SMController. S'ha definit 'Redirect' a nivell de classe (" + classRedirect.value() + ").");
            response.sendRedirect(request.getContextPath() + (classRedirect.value().startsWith("/") ? classRedirect.value() : ("/" + classRedirect.value())));
        } else if (classForward != null) {
            debug("SMController. S'ha definit 'Forward' a nivell de classe (" + methodForward.value() + ").");
            request.getRequestDispatcher(classForward.value().startsWith("/") ? classForward.value() : ("/" + classForward.value())).forward(request, response);
        }
    }

    
    /**
     * Analitza un atribut d'una Action i tracta de recuperar la informació del
     * request per donar valor a l'atritbut.
     *
     * @param request Peticio HTTP
     * @param field Camp de la classe Action marcat com SMControllerBean
     * @return instancia del tipus field.getType() amb el valor rebut pel request
     * correctament mapejat
     */
    private Object getParametersFromRequest(HttpServletRequest request, Field field) {
        SMControllerBean annotation = field.getAnnotation(SMControllerBean.class);

        if (annotation.requestGetParameter()) {
            debug("Mapeig del parametre " + field.getName() + " de tipus " + field.getType() + " a l'action " + field.getDeclaringClass().getName() + " a partir de les dades del request");
        }

        Object obj = null;
        if (annotation.requestGetParameter() && Data.class.isAssignableFrom(field.getType())) {
            debug("Realitzant mapeig " + field.getName() + " com a classe Data");
            try {
                obj = field.getType().newInstance();
                Data data = (Data) obj;
                BeanUtilsBean instance = getBeanUtilsBeanInstance();
                instance.populate(data, request.getParameterMap());
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipus Data, pero no es possible realitzar la conversio: " + e.getMessage());
            }
        } else if (annotation.requestGetParameter() && Integer.class.equals(field.getType())) {
            debug("Realitzant mapeig " + field.getName() + " com a classe Integer");
            Integer i = Integer.valueOf(0);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    i = Integer.parseInt(s);
                }
                obj = i;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipus Integer, pero no es possible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && Float.class.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " com classe Float");
            Float f = Float.valueOf(0.0f);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    f = Float.parseFloat(s);
                }
                obj = f;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipusFloat, pero no es posible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && Double.class.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " com classe Double");
            Double d = Double.valueOf(0.0d);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    d = Double.parseDouble(s);
                }
                obj = s;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipusDouble, pero no es posible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && Integer.TYPE.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " como classe Integer");
            Integer i = Integer.valueOf(0);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    i = Integer.parseInt(s);
                }
                obj = i;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipusInteger, pero no es posible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && Float.TYPE.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " com classe Float");
            Float f = Float.valueOf(0.0f);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    f = Float.parseFloat(s);
                }
                obj = f;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la classe " + field.getType() + "se va declarar de tipusFloat, pero no es posible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && Double.TYPE.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " com classe Double");
            Double d = Double.valueOf(0.0d);
            String s = request.getParameter(field.getName());
            try {
                if (s != null && !s.equals(Constants.EMPTY_STRING)) {
                    d = Double.parseDouble(s);
                }
                obj = s;
            } catch (Exception e) {
                debug("El parametre " + field.getName() + " de la clase " + field.getType() + "se va declarar de tipusDouble, pero no es posible realitzar la conversio.");
            }
        } else if (annotation.requestGetParameter() && String.class.equals(field.getType())) {
            debug("Realitzant mapeig" + field.getName() + " como clase String");
            String s = request.getParameter(field.getName());
            obj = s;
        } else if (annotation.requestGetParameter()) {
            debug("S'ha sol licitat el mapeig dels parametres del request a la instancia " + field.getName() + " de la classe " + field.getType() + " pero no es posible realitzar-lo perque no es d'un tipus suportat per SMFramework");        }

        return obj;
    }

    /**
     * Asigna el valor value a l'atribut field d'una instancia instance.
     *
     * @param field Atribut a assignar
     * @param instance Instancia a la que hem d'assignar el valor
     * @param value Valor a assignar
     */
    private void setFieldValue(Field field, Object instance, Object value) {
        try {
            
            debug("SMController. Processament de l'assignació de valor a un atritbut (field) ");
            
            if (field == null || instance == null || value == null) {
                if (field == null) {                    
                    debug(" - Intentant assignar valor a un atribut nul. No s'assigna res.");
                }
                if (instance == null) {
                    debug(" - Intentant assignar valor a una instancia nula. No s'assigna res.");
                }
                if (value == null) {
                    debug(" - Intentant assignar valor nul a un attribut" + (field != null ? " " + field.getName() : "") + ". No s'assigna res.");
                }
                return;
            }

            debug("SMController. Assignant valor per reflection a " + instance.getClass().getName() + "." + field.getName() + " amb el valor " + value);

            boolean accessible = field.isAccessible();
            field.setAccessible(true);
            field.set(instance, value);
            field.setAccessible(accessible);
        } catch (Exception ex) {
            debug("SMController. No es possible assignar el valor " + value + " a l'atribut " + field.getName() + " de la classe " + instance.getClass().getName() + ": " + SMControllerError.stack2string(ex));
        }
    }
    
    
    

    /**
     * Crea una instancia d'objecte de la classe 'type' emprant els constructors
     * mes coneguts, començant pel constructor buit, seguint pel numeric i 
     * finalitzant per la cadena buida.
     * 
     * @param type Classe de la que se vol crear la instancia
     * @return Instancia de l'objecto creat
     */
    private <A> A createNewInstance(Class<A> type) {
        if (type == null) {
            return null;
        }
        try {
            Constructor constructor = type.getDeclaredConstructor((Class[]) null);
            return (A) constructor.newInstance((Object[]) null);
        } catch (Exception ex) {
        }

        try {
            Constructor constructor = type.getDeclaredConstructor(String.class);
            return (A) constructor.newInstance(Constants.EMPTY_STRING);
        } catch (Exception ex) {
        }

        try {
            Constructor constructor = type.getDeclaredConstructor(String.class);
            return (A) constructor.newInstance("0");
        } catch (Exception ex) {
        }

        try {
            Constructor constructor = type.getDeclaredConstructor(Integer.TYPE);
            return (A) constructor.newInstance(0);
        } catch (Exception ex) {
        }
        return null;
    }
    
    
    /**
     * Obté un valor sencer per parametre del request.
     * 
     * @param request el request
     * @param name nom del parametre del que volem extreure un sencer
     * @return el nombre sencer.
     */
    private int getIntFromRequest(HttpServletRequest request, String name) {
        try {
            return Integer.parseInt(request.getParameter(name));
        } catch (Exception e) {
            return Constants.DEFAULT_INT;
        }
    }
    
    /**
     * Obté un valor float per parametre del request.
     * 
     * @param request el request
     * @param name nom del parametre del que volem extreure el float
     * @return el valor float.
     */
    private float getFloatFromRequest(HttpServletRequest request, String name) {
        try {
            return Float.parseFloat(request.getParameter(name));
        } catch (Exception e) {
            return Constants.DEFAULT_FLOAT;
        }
    }
    
    /**
     * Obté un valor double per parametre del request.
     * 
     * @param request el request
     * @param name nom del parametre del que volem extreure un double
     * @return el valor double.
     */
    private double getDoubleFromRequest(HttpServletRequest request, String name) {
        try {
            return Double.parseDouble(request.getParameter(name));
        } catch (Exception e) {
            return Constants.DEFAULT_DOUBLE;
        }
    }
    
    
    /**
     * Obté un valor char per parametre del request.
     * 
     * @param request el request
     * @param name nom del parametre del que volem extreure un el char
     * @return el valor char.
     */
    private char getCharFromRequest(HttpServletRequest request, String name) {
        try {
            return request.getParameter(name).charAt(0);
        } catch (Exception e) {
            return Constants.DEFAULT_CHAR;
        }
    }
    
    /**
     * Obté un valor boolea per parametre del request.
     * 
     * @param request el request
     * @param name nom del parametre del que volem extreure el boolea
     * @return el valor boolea.
     */
    private boolean getBooleanFromRequest(HttpServletRequest request, String name) {
        try {
            String s = request.getParameter(name);
            if ("S".equals(s)) {
                return true;
            } else if ("Y".equals(s)) {
                return true;
            } else if ("true".equals(s)) {
                return true;
            } else if ("on".equals(s)) {
                return true;
            } else if (Integer.parseInt(s) == 1) {
                return true;
            } else {
                return false;
            }
        } catch (Exception e) {
            return Constants.DEFAULT_BOOLEAN;
        }
    }
    
    
    /**
     * Registra conversors per convertir valors de String segons la classe que 
     * els correspon fent servir les funcionalitats de Apache Commons BeanUtils.
     *
     * @return Instancia de BeanUtilsBean amb els convertidors definits.
     */
    private BeanUtilsBean getBeanUtilsBeanInstance() {
        BeanUtilsBean instance = BeanUtilsBean2.getInstance();
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.DateConverter.getInstance(), java.util.Date.class);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.TimestampConverter.getInstance(), java.sql.Timestamp.class);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Integer.class);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Integer.TYPE);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Float.class);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Float.TYPE);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Double.class);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), Double.TYPE);
        instance.getConvertUtils().register(edu.uoc.pfc.sm.beanutils.converter.NumberConverter.getInstance(), java.math.BigDecimal.class);
        return instance;
    }
    
    
    /**
     * Si esta activada l'opcio de debug, imprimeix el missatge debugMessage
     * a la sortida estandard.
     * 
     * @param debugMessage 
     */
    private void debug(String debugMessage) {
        if (this.debug) {
            System.out.println(debugMessage);
        }
    }
    
    
        /**
     * Gestiona el metode HTTP <code>GET</code>.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /**
     * Gestiona el metode HTTP <code>POST</code>.     
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /**
     * Retorna una breu descripció del Servlet
     *
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "SMController 1.0";
    }
}
