package edu.uoc.pfc.sm.validation;

import edu.uoc.pfc.sm.annotation.FieldValidation;
import edu.uoc.pfc.sm.annotation.SMControllerBean;
import edu.uoc.pfc.sm.annotation.ValidateBean;
import edu.uoc.pfc.sm.annotation.ValidateMethod;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import javax.servlet.http.HttpServletRequest;

/**
 * Classe que processa la validació de dades d'entrada de la vista per al 
 * controlador SMController.
 * 
 * @author Salvador Morla
 * @version 1.0
 */
public class SMControllerValidator {
    
    private boolean debug = false;
    private String  operationName;
    private Object  action;
    private HttpServletRequest request;
    
    private SMControllerValidationError error;
    
    public SMControllerValidator() {
        error = new SMControllerValidationError();
    }

    public SMControllerValidator(Object action, String operationName, HttpServletRequest request, boolean debug) {
        
        this.action = action;
        this.operationName = operationName;
        this.request = request;
        this.debug = debug;        
        this.error = new SMControllerValidationError();
    }
    
    
   /**
    * Processament de la validació de dades d'entrada.
    * Se recorren els atributs dels SMControllerBeans i se procesen les 
    * entrades que han estat anotades amb validate.
    * La validació se realitza a nivell de metode. 
    */
    public void executeValidation() throws Exception {
        
        try {            
            Method[] declaredMethods = action.getClass().getDeclaredMethods();
            for (Method method : declaredMethods) {
                
                // Efectuam la validació pel metode de la classe Action si
                // aquest metode es el que s'ha indicat a la peticio
                if (operationName.equals(method.getName()) && method.getAnnotation(ValidateMethod.class) != null) { 
                    debug("SMController. Validació dels SMControllerBeans per al metode " + operationName + ".");
                    
                    // Recuperam de el valor de la annotation que indica la vista a mostrar
                    // en cas de que se produeixi un error de validació.
                    ValidateMethod vm = method.getAnnotation(ValidateMethod.class);
                    getError().setValidationErrorView(vm.errorView());
                    
                    Field[] declaredFields = action.getClass().getDeclaredFields();
                    for (Field f: declaredFields) {                        
                        if (f.getAnnotation(SMControllerBean.class) != null
                            && f.getAnnotation(ValidateBean.class) != null) {                            
                            try {
                                String param;
                                Field[] properties = f.getType().getDeclaredFields();
                                for (Field p: properties) {
                                    if (p.getAnnotation(FieldValidation.class) != null) {
                                        param = request.getParameter(p.getName());
                                        debug("Validació del camp "+p.getName()+" amb valor = "+param);
                                        SMControllerFieldError fe = validate(param, p);                                        
                                        if (fe != null) {
                                            getError().getErrorList().add(fe);
                                        }
                                    }
                                }                                    
                            }
                            catch (Exception ex) {
                                debug("SMControllerBean. Error intern processant la valicació de camps. No se continua amb la validació. ");
                            }
                        }
                    }                        
                }                                        
            }                            
        } 
        catch (Exception ex) {
            throw ex;
        }        
    }
    
    /**
     * Valida un camp d'un formulari que arriba al request contra
     * la validació definida a l'annotacio indicada a la classe que l'inclou
     * 
     * @param param el valor del camp en el formulari
     * @param field el field amb la annotacio que especifica la seva validacio
     * @return retorna null si el parametre es camp valida correctament i l'error
     * en cas contrari.
     */
    private SMControllerFieldError validate (String param, Field field) throws Exception {
        
        // Declaracions
        SMControllerFieldError fe = null;
        
        // Obtenim els parametres de la annotació Validate pel camp.
        FieldValidation v = field.getAnnotation(FieldValidation.class);
        Object validationClass = null;
        try {            
            // Veim si se tracta d'un dels tipus de validació suportats al 
            // SMFramework directament
            validationClass = Thread.currentThread().getContextClassLoader().loadClass("edu.uoc.pfc.sm.validation." + v.type()).newInstance();
        } catch (Exception ex) {
            // En aquest cas se tracta d'una valació propia definida per l'usuari
            debug("SMController. Validacio customitzada d'usuari ("+v.type()+").");
            try {
                validationClass = Thread.currentThread().getContextClassLoader().loadClass(v.type()).newInstance();            
            }
            catch (Exception exx) {
                debug("SMController. No se pot validar: no s'ha trobat la implementacio de la classe de validació indicada ("+v.type()+").");
                return null;
            }            
        }
        
        // Hem d'especificar els valors dels parametres de validació definits
        // a l'anotació del 'field'
        ((SMControllerValidationType)validationClass).setValue(param);
        ((SMControllerValidationType)validationClass).setPattern(v.pattern());
        ((SMControllerValidationType)validationClass).setMaxLength(v.maxLength());
        ((SMControllerValidationType)validationClass).setMinLength(v.minLength());
        ((SMControllerValidationType)validationClass).setMaxValue(v.maxValue());
        ((SMControllerValidationType)validationClass).setMinValue(v.minValue());
        ((SMControllerValidationType)validationClass).setNotNull(v.notNull());
        
        // Execució de la validació.
        Method m = validationClass.getClass().getMethod("validate", (Class<?>[]) null);
        fe = (SMControllerFieldError)m.invoke(validationClass, (Object[]) null);
        
        // Si hi ha error indicam el nom del camp que no ha superat la validacio
        if (fe != null) {
            fe.setFieldName(field.getName());
        }
        debug("SMController. El camp "+ field.getName() + ((fe != null)?" no":"") + " ha superat la validació.");
        return fe;
    }
    
    
    /**
     * Si esta activada l'opcio de debug, imprimeix el missatge debugMessage
     * a la sortida estandard.
     * 
     * @param debugMessage 
     */
    private void debug(String debugMessage) {
        if (this.debug) {
            System.out.println(debugMessage);
        }
    }

    
    /**
     * Getter per la propietat error
     * @return el valor de la propietat
     */
    public SMControllerValidationError getError() {
        return error;
    }    
    
}
